from classicML.api.models.linear_model.linear_discriminant_analysis import LinearDiscriminantAnalysis
from classicML.api.models.linear_model.linear_discriminant_analysis import LinearDiscriminantAnalysis as LDA
from classicML.api.models.linear_model.logistic_regression import LogisticRegression

from classicML.api.models.neural_network.back_propagation_neural_network import BackPropagationNeuralNetwork
from classicML.api.models.neural_network.back_propagation_neural_network import BackPropagationNeuralNetwork as BPNN
from classicML.api.models.neural_network.radial_basis_function_network import RadialBasisFunctionNetwork
from classicML.api.models.neural_network.radial_basis_function_network import RadialBasisFunctionNetwork as RBF

from classicML.api.models.support_vector_machine.support_vector_classifier import SupportVectorClassifier
from classicML.api.models.support_vector_machine.support_vector_classifier import SupportVectorClassifier as SVC

from classicML.api.models.tree.decision_tree_classifier import DecisionTreeClassifier
